if (not Grid) then return end

-- Libraries

local L = AceLibrary("AceLocale-2.2"):new("GridLayoutPlus")
local RL = AceLibrary("Roster-2.1")

-- New Grid module

GridLayoutPlus = Grid:NewModule("GridLayoutPlus", "AceConsole-2.0")

GridLayoutPlus.defaultDB = {
	autoLayout = true,
	layoutSorting = "By Group",
	withPets = true,
	withTanks = false,
	maxColumns = 18,
	warnPetsHidden = false,
	warnTanksHidden = false,
--	classOrder = {
--		[1] = "WARRIOR",
--		[2] = "ROGUE",
--		[3] = "DRUID",
--		[4] = "PALADIN",
--		[5] = "SHAMAN",
--		[6] = "PRIEST",
--		[7] = "MAGE",
--		[8] = "WARLOCK",
--		[9] = "HUNTER",
--	},
--	petOrder = {
--		[1] = "HUNTER",
--		[2] = "WARLOCK",
--		[3] = "MAGE",
--		[4] = "PRIEST",
--	},
--	roleOrder = {
--		[1] = "MAINTANK",
--		[2] = "MAINASSIST",
--	},
}


-- Locals

local classOrder = "WARRIOR,ROGUE,DRUID,PALADIN,SHAMAN,PRIEST,MAGE,WARLOCK,HUNTER"
local petOrder = "HUNTER,WARLOCK"
local roleOrder = "MAINTANK,MAINASSIST"

local playerColumns = 0
local petColumns = 0
local tankColumns = 0
local tankType = nil
local raTankList = nil

local updateNeeded = nil


-- Default layout

GridLayoutPlus.DefaultLayout = {
	defaults = {
		unitsPerColumn = 5,
		maxColumns = 8,
	},
	[1] = nil,
	[2] = nil,
	[3] = nil,
	[4] = nil,
	[5] = nil,
	[6] = nil,
	[7] = nil,
	[8] = nil,
	[9] = nil,
	[10] = nil,
}

GridLayout:AddLayout(L["GridLayoutPlus"], GridLayoutPlus.DefaultLayout)


-- Init

function GridLayoutPlus:OnEnable()
	local options = GridLayout.options.args
	local ORDER_PLUS = 40
	options["PlusHeader"] = {
		type = "header",
		order = ORDER_PLUS,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["autoLayout"] = {
		type = "toggle",
		name = L["Auto adjust layout (+)"],
		desc = L["Automatically adjust layout to fit raid size"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.autoLayout
		end,
		set = function (v)
			GridLayoutPlus.db.profile.autoLayout = v
			GridLayoutPlus:FullUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["layoutSorting"] = {
		type = "text",
		name = L["Layout sorting (+)"],
		desc = L["Choose between By Group or By Class"],
		order = ORDER_PLUS + 1,
		get = function ()
			return GridLayoutPlus.db.profile.layoutSorting
		end,
		set = function (v)
			GridLayoutPlus.db.profile.layoutSorting = v
			GridLayoutPlus:RosterUpdate()
		end,
		validate = {
			["By Group"] = L["By Group"],
			["By Class"] = L["By Class"],
		},
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["withPets"] = {
		type = "toggle",
		name = L["Use pet layouts (+)"],
		desc = L["Use pet layouts when auto adjusting (if possible)"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.withPets
		end,
		set = function (v)
			GridLayoutPlus.db.profile.withPets = v
			GridLayoutPlus:RosterUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["withTanks"] = {
		type = "toggle",
		name = L["Use tank layouts (+)"],
		desc = L["Use tank layouts when auto adjusting (if possible)"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.withTanks
		end,
		set = function (v)
			GridLayoutPlus.db.profile.withTanks = v
			GridLayoutPlus:FullUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["PlusAdvanced"] = {
		type = "group",
		name = L["Plus Advanced (+)"],
		desc = L["Advanced options for GridLayoutPlus"],
		order = ORDER_PLUS,
		args = {
			["maxColumns"] = {
				type = "range",
				name = L["Max number of columns"],
				desc = L["Maximum number of columns when auto-adjusting (WARNING: May hide units)"],
				order = 1,
				min = 5,
				max = 18,
				step = 1,
				get = function()
					return GridLayoutPlus.db.profile.maxColumns
				end,
				set = function(v)
					GridLayoutPlus.db.profile.maxColumns = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
			["warnPetsHidden"] = {
				type = "toggle",
				name = L["Warn if pets are hidden"],
				desc = L["Show warning when pet units are forced to be hidden"],
				order = 4,
				get = function()
					return GridLayoutPlus.db.profile.warnPetsHidden
				end,
				set = function(v)
					GridLayoutPlus.db.profile.warnPetsHidden = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
			["warnTanksHidden"] = {
				type = "toggle",
				name = L["Warn if tanks are hidden"],
				desc = L["Show warning when tank units are forced to be hidden"],
				order = 5,
				get = function()
					return GridLayoutPlus.db.profile.warnTanksHidden
				end,
				set = function(v)
					GridLayoutPlus.db.profile.warnTanksHidden = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
--			["classOrder"] = {
--				type = "text",
--				name = L["Class order"],
--				desc = L["Order in which player units appear"],
--				order = 6,
--				get = function()
--					return GridLayoutPlus:GenerateOrder(GridLayoutPlus.db.profile.classOrder)
--				end,
--				set = function(v)
--					GridLayoutPlus.db.profile.classOrder = GridLayoutPlus:ParseOrder(v)
--					updateNeeded = true
--					GridLayoutPlus:UpdateLayout()
--				end,
--			},
--			["petOrder"] = {
--				type = "text",
--				name = L["Pet order"],
--				desc = L["Order in which pet units appear"],
--				order = 7,
--				get = function()
--					return GridLayoutPlus:GenerateOrder(GridLayoutPlus.db.profile.petOrder)
--				end,
--				set = function(v)
--					GridLayoutPlus.db.profile.petOrder = GridLayoutPlus:ParseOrder(v)
--					updateNeeded = true
--					GridLayoutPlus:UpdateLayout()
--				end,
--			},
--			["roleOrder"] = {
--				type = "text",
--				name = L["Blizzard-tank order"],
--				desc = L["Order in which blizzard tanks appear"],
--				order = 8,
--				get = function()
--					return GridLayoutPlus:GenerateOrder(GridLayoutPlus.db.profile.roleOrder)
--				end,
--				set = function(v)
--					GridLayoutPlus.db.profile.roleOrder = GridLayoutPlus:ParseOrder(v)
--					updateNeeded = true
--					GridLayoutPlus:UpdateLayout()
--				end,
--			},
		},
	}
	self:RegisterEvent("Grid_Enabled", "FullUpdate")
	self:RegisterEvent("Grid_JoinedRaid", "FullUpdate")
	self:RegisterEvent("Grid_JoinedBattleground", "FullUpdate")
	self:RegisterEvent("Grid_UnitJoined", "RosterUpdate")
	self:RegisterEvent("Grid_UnitLeft", "RosterUpdate")
	self:RegisterEvent("Grid_UnitChanged", "UnitUpdate")
	self:RegisterEvent("Grid_LeftParty", "ClearLocals")
	self:RegisterEvent("Grid_Disabled", "ClearLocals")
	if oRA then
		self:RegisterEvent("oRA_MainTankUpdate", "RATanksUpdate")
		self:RegisterEvent("oRA_JoinedRaid", "RATanksUpdate")
	elseif CT_RA_MainTanks then
		hooksecurefunc("CT_RAOptions_UpdateMTs", function() self:RATanksUpdate() end)
	end
	self:RegisterEvent("GridLayoutPlus_UpdateLayout", "UpdateLayout")
	self:ScheduleRepeatingEvent("GridLayoutPlus_UpdateLayout", 5)
	self:FullUpdate()
end


-- Utility functions

function GridLayoutPlus:ParseOrder(str)
	local result = {}
	local o = 1
	if string.find(str, ",") then
		for a in gmatch(str..",", "(.-),") do
			if a ~= "" then
				result[o] = a
			end
			o = o + 1
		end
	elseif s ~= "" then
		result[o] = str
	end
	table.sort(result)
	return result
end

function GridLayoutPlus:GenerateOrder(list)
	local result = ""
	for key,val in pairs(list) do
		result = result .. val .. ","
	end
	return result
end


-- Event handlers

function GridLayoutPlus:FullUpdate()
	self:RATanksUpdate()
	self:RosterUpdate()
	updateNeeded = true
	self:UpdateLayout()
end

function GridLayoutPlus:UnitUpdate(name, unitid)
	local u = RL:GetUnitObjectFromName(name)
	if u and (u.class ~= "PET") then
		if (GridLayoutPlus.db.profile.layoutSorting == "By Group")
				or (GridLayoutPlus.db.profile.withTanks and (((tankType ~= "RA") and ((u.role == "MAINTANK") or (u.role == "MAINASSIST")))
														or ((tankType == "Blizz") and (u.role ~= "MAINTANK") and (u.role ~= "MAINASSIST")))) then
			self:RosterUpdate()
		end
	end
end

function GridLayoutPlus:RATanksUpdate()
	local ratanktable = nil
	local numberoftanks = 0
	if GridLayoutPlus.db.profile.withTanks and oRA and oRA.maintanktable then
		ratanktable = oRA.maintanktable
	elseif GridLayoutPlus.db.profile.withTanks and CT_RA_MainTanks then
		ratanktable = CT_RA_MainTanks
	end
	if ratanktable then
		for key, val in pairs(ratanktable) do
			local unitid = RL:GetUnitIDFromName(val)
			if unitid and UnitExists(unitid) and UnitPlayerOrPetInRaid(unitid) then
				numberoftanks = numberoftanks + 1
			end
		end
	end
	tankColumns = floor((numberoftanks / 5) + .8)
	if (tankColumns > 0) then
		tankType = "RA"
		if oRA and oRA.maintanktable then
			raTankList = table.concat(oRA.maintanktable, ',')
		elseif CT_RA_MainTanks then
			raTankList = table.concat(CT_RA_MainTanks, ',')
		end
	elseif (tankType == "RA") then
		tankType = nil
		raTankList = nil
	end
	updateNeeded = true
end

function GridLayoutPlus:RosterUpdate()
	local maxsubgroup = 0
	local numberofpotentialpets = 0
	local numberoftanks = 0
	for u in RL:IterateRoster(nil) do
		if (GridLayoutPlus.db.profile.layoutSorting == "By Group") and (u.subgroup > maxsubgroup) then
			maxsubgroup = u.subgroup
		end
		if GridLayoutPlus.db.profile.withPets and ((u.class == "HUNTER") or (u.class == "WARLOCK")) then
			numberofpotentialpets = numberofpotentialpets + 1
		end
		if GridLayoutPlus.db.profile.withTanks and (tankType ~= "RA") and ((u.role == "MAINTANK") or (u.role == "MAINASSIST")) then
			numberoftanks = numberoftanks + 1
		end
	end
	if (GridLayoutPlus.db.profile.layoutSorting == "By Group") and (playerColumns ~= maxsubgroup) then
		playerColumns = maxsubgroup
		updateNeeded = true
	elseif (playerColumns ~= floor((GetNumRaidMembers() / 5) + .8)) then
		playerColumns = floor((GetNumRaidMembers() / 5) + .8)
		updateNeeded = true
	end
	if (petColumns ~= floor((numberofpotentialpets / 5) + .8)) then
		petColumns = floor((numberofpotentialpets / 5) + .8)
		updateNeeded = true
	end
	if (tankColumns ~= floor((numberoftanks / 5) + .8)) then
		tankColumns = floor((numberoftanks / 5) + .8)
		if (tankColumns > 0) then
			tankType = "Blizz"
		elseif (tankType == "Blizz") then
			tankType = nil
		end
		updateNeeded = true
	end
end

function GridLayoutPlus:ClearLocals()
	playerColumns = 0
	petColumns = 0
	tankColumns = 0
	tankType = nil
	raTankList = nil
end


-- Update layout

function GridLayoutPlus:UpdateLayout()
	if (updateNeeded) then
		if GridLayoutPlus.db.profile.autoLayout and (GetNumRaidMembers() > 0) and (not InCombatLockdown()) then
			local canfitpets = nil
			local canfittanks = nil
			local groupindex = 1
			-- Can fit pets / tanks
			if (GridLayoutPlus.db.profile.maxColumns >= (playerColumns + petColumns + tankColumns)) then
				canfitpets = true
				canfittanks = true
			elseif (GridLayoutPlus.db.profile.maxColumns >= (playerColumns + petColumns)) then
				canfitpets = true
				canfittanks = nil
			else
				canfitpets = nil
				canfittanks = nil
			end
			-- Alter layout
			if GridLayoutPlus.db.profile.withTanks and canfittanks and (tankType == "RA") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					nameList = raTankList,
					sortMethod = "INDEX",
				}
				groupindex = groupindex + 1
			elseif GridLayoutPlus.db.profile.withTanks and canfittanks and (tankType == "Blizz") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = roleOrder,
					groupingOrder = roleOrder,
					groupBy = "ROLE",
				}
				groupindex = groupindex + 1
			end
			if (GridLayoutPlus.db.profile.layoutSorting == "By Group") then
				for groupnumber = 1, playerColumns do
					GridLayoutPlus.DefaultLayout[groupindex] = {
						groupFilter = groupnumber,
						groupingOrder = classOrder,
						groupBy = "CLASS",
					}
					groupindex = groupindex + 1
				end
			elseif (GridLayoutPlus.db.profile.layoutSorting == "By Class") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = classOrder,
					groupingOrder = classOrder,
					groupBy = "CLASS",
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.withPets and canfitpets and (petColumns > 0) then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					isPetGroup = true,
					groupFilter = petOrder,
					groupingOrder = petOrder,
					groupBy = "CLASS",
					filterOnPet = true,
				}
				groupindex = groupindex + 1
			end
			for i = groupindex, 10 do
				if GridLayoutPlus.DefaultLayout[i] then
					GridLayoutPlus.DefaultLayout[i] = nil
				end
			end
			-- Refresh layout
			if GridLayout.db.profile.layout ~= "GridLayoutPlus" then
				GridLayout.db.profile.layout = "GridLayoutPlus"
				GridLayout:LoadLayout("GridLayoutPlus")
			else
				GridLayout:ReloadLayout()
			end
			-- Warn if hidden units
			if GridLayoutPlus.db.profile.warnPetsHidden and GridLayoutPlus.db.profile.withPets and (petColumns > 0) and (not canfitpets) then
				self:Print(L["Your maxColumns setting is causing pet units to be hidden"])
			end
			if GridLayoutPlus.db.profile.warnTanksHidden and GridLayoutPlus.db.profile.withTanks and (tankColumns > 0) and (not canfittanks) then
				self:Print(L["Your maxColumns setting is causing tank units to be hidden"])
			end
			updateNeeded = nil
		elseif GridLayoutPlus.db.profile.autoLayout and (GetNumRaidMembers() > 0) and InCombatLockdown() then
			self:ScheduleLeaveCombatAction(self, "RosterUpdate")
		end
	end
end
